
import json
import os
import shutil
import tempfile
from typing import Literal
from fastapi import Request

import httpx

from olah.constants import CHUNK_SIZE, WORKER_API_TIMEOUT
from olah.utls import check_cache_rules_hf

async def file_head_generator(app, repo_type: Literal["model", "dataset"], org: str, repo: str, commit: str, file_path: str, request: Request):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    # save
    repos_path = app.app_settings.repos_path
    save_path = os.path.join(repos_path, f"heads/{repo_type}s/{org}/{repo}/resolve_head/{commit}/{file_path}")
    save_dir = os.path.dirname(save_path)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)
    
    use_cache = os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    # proxy
    if use_cache:
        with open(save_path, "r", encoding="utf-8") as f:
            response_headers = json.loads(f.read())
            if "location" in response_headers:
                response_headers["location"] = response_headers["location"].replace(app.app_settings.hf_lfs_url, app.app_settings.mirror_lfs_url)
            yield response_headers
    else:
        if repo_type == "model":
            url = f"{app.app_settings.hf_url}/{org}/{repo}/resolve/{commit}/{file_path}"
        else:
            url = f"{app.app_settings.hf_url}/{repo_type}s/{org}/{repo}/resolve/{commit}/{file_path}"
        async with httpx.AsyncClient() as client:
            async with client.stream(
                method="HEAD", url=url,
                headers=headers,
                timeout=WORKER_API_TIMEOUT,
            ) as response:
                response_headers = response.headers
                response_headers = {k: v for k, v in response_headers.items()}
                if allow_cache:
                    with open(save_path, "w", encoding="utf-8") as f:
                        f.write(json.dumps(response_headers, ensure_ascii=False))
                if "location" in response_headers:
                    response_headers["location"] = response_headers["location"].replace(app.app_settings.hf_lfs_url, app.app_settings.mirror_lfs_url)
                yield response_headers
                
                async for raw_chunk in response.aiter_raw():
                    if not raw_chunk:
                        continue 
                    yield raw_chunk


async def file_get_generator(app, repo_type: Literal["model", "dataset"], org: str, repo: str, commit: str, file_path: str, request: Request):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")
    # save
    repos_path = app.app_settings.repos_path
    save_path = os.path.join(repos_path, f"files/{repo_type}s/{org}/{repo}/resolve/{commit}/{file_path}")
    save_dir = os.path.dirname(save_path)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)
    
    use_cache = os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    # proxy
    if use_cache:
        yield request.headers
        with open(save_path, "rb") as f:
            while True:
                chunk = f.read(CHUNK_SIZE)
                if not chunk:
                    break
                yield chunk
    else:
        try:
            temp_file_path = None
            if repo_type == "model":
                url = f"{app.app_settings.hf_url}/{org}/{repo}/resolve/{commit}/{file_path}"
            else:
                url = f"{app.app_settings.hf_url}/{repo_type}s/{org}/{repo}/resolve/{commit}/{file_path}"
            async with httpx.AsyncClient() as client:
                with tempfile.NamedTemporaryFile(mode="wb", delete=False) as temp_file:
                    if not allow_cache:
                        temp_file = open(os.devnull, 'wb')
                    async with client.stream(
                        method="GET", url=url,
                        headers=headers,
                        timeout=WORKER_API_TIMEOUT,
                    ) as response:
                        response_headers = response.headers
                        yield response_headers

                        async for raw_chunk in response.aiter_raw():
                            if not raw_chunk:
                                continue
                            temp_file.write(raw_chunk)
                            yield raw_chunk
                    if not allow_cache:
                        temp_file_path = None
                    else:
                        temp_file_path = temp_file.name
                if temp_file_path is not None:
                    shutil.copyfile(temp_file_path, save_path)
        finally:
            if temp_file_path is not None:
                os.remove(temp_file_path)





from typing import List, Optional
import toml
import re
import fnmatch

DEFAULT_PROXY_RULES = [
    {
        "repo": "*/*",
        "allow": True,
        "use_re": False
    }
]

DEFAULT_CACHE_RULES = [
    {
        "repo": "*/*",
        "allow": True,
        "use_re": False
    }
]

class OlahRule(object):
    def __init__(self) -> None:
        self.repo = ""
        self.type = "*"
        self.allow = False
        self.use_re = False

    @staticmethod
    def from_dict(data) -> "OlahRule":
        out = OlahRule()
        if "repo" in data:
            out.repo = data["repo"]
        if "allow" in data:
            out.allow = data["allow"]
        if "use_re" in data:
            out.use_re = data["use_re"]
        return out
    
    def match(self, repo_name: str) -> bool:
        if self.use_re:
            return self.match_re(repo_name)
        else:
            return self.match_fn(repo_name)
    
    def match_fn(self, repo_name: str) -> bool:
        return fnmatch.fnmatch(repo_name, self.repo)
    
    def match_re(self, repo_name: str) -> bool:
        return re.match(self.repo, repo_name) is not None

class OlahRuleList(object):
    def __init__(self) -> None:
        self.rules: List[OlahRule] = []
    
    @staticmethod
    def from_list(data) -> "OlahRuleList":
        out = OlahRuleList()
        for item in data:
            out.rules.append(OlahRule.from_dict(item))
        return out
    
    def clear(self):
        self.rules.clear()
    
    def allow(self, repo_name: str) -> bool:
        allow = False
        for rule in self.rules:
            if rule.match(repo_name):
                allow = rule.allow
        return allow

class OlahConfig(object):
    def __init__(self, path: Optional[str] = None) -> None:

        # basic
        self.host = "localhost"
        self.port = 8090
        self.ssl_key = None
        self.ssl_cert = None
        self.repos_path = "./repos"
        self.hf_url = "https://huggingface.co"
        self.hf_lfs_url = "https://cdn-lfs.huggingface.co"
        self.mirror_url = "http://localhost:8090"
        self.mirror_lfs_url = "http://localhost:8090"

        # accessibility
        self.proxy = OlahRuleList.from_list(DEFAULT_PROXY_RULES)
        self.cache = OlahRuleList.from_list(DEFAULT_CACHE_RULES)

        if path is not None:
            self.proxy.clear()
            self.cache.clear()
            self.read_toml(path)
    
    def empty_str(self, s: str) -> Optional[str]:
        if s == "":
            return None
        else:
            return s

    def read_toml(self, path: str):
        config = toml.load(path)

        basic = config["basic"]
        accessibility = config["accessibility"]

        self.host = basic["host"]
        self.port = basic["port"]
        self.ssl_key = self.empty_str(basic["ssl-key"])
        self.ssl_cert = self.empty_str(basic["ssl-cert"])
        self.repos_path = basic["repos-path"]
        self.hf_url = basic["hf-url"]
        self.hf_lfs_url = basic["hf-lfs-url"]
        self.mirror_url = basic["mirror-url"]
        self.mirror_lfs_url = basic["mirror-lfs-url"]

        self.proxy = OlahRuleList.from_list(accessibility["proxy"])
        self.cache = OlahRuleList.from_list(accessibility["cache"])



import datetime
import json
import os
import argparse
import tempfile
import shutil
from typing import Annotated, Union
from fastapi import FastAPI, Header, Request
from fastapi.responses import StreamingResponse, Response
import httpx
from pydantic import BaseSettings
import pytz
from olah.configs import OlahConfig
from olah.files import file_get_generator, file_head_generator
from olah.lfs import lfs_get_generator
from olah.meta import meta_generator
from olah.utls import check_proxy_rules_hf, check_commit_hf

app = FastAPI(debug=False)

class AppSettings(BaseSettings):
    # The address of the model controller.
    config: OlahConfig = OlahConfig()
    repos_path: str = "./repos"
    hf_url: str = "https://huggingface.co"
    hf_lfs_url: str = "https://cdn-lfs.huggingface.co"
    mirror_url: str = "http://localhost:8090"
    mirror_lfs_url: str = "http://localhost:8090"


@app.get("/api/{repo_type}s/{org}/{repo}/revision/{commit}")
async def meta_proxy(repo_type: str, org: str, repo: str, commit: str, request: Request):
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return Response(content="This repository is forbidden by the mirror. ", status_code=403)
    if not await check_commit_hf(app, repo_type, org, repo, commit):
        return Response(content="This repository is not accessible. ", status_code=404)
    generator = meta_generator(app, repo_type, org, repo, commit, request)
    headers = await generator.__anext__()
    return StreamingResponse(generator, headers=headers)

@app.head("/{repo_type}s/{org}/{repo}/resolve/{commit}/{file_path:path}")
@app.head("/{org}/{repo}/resolve/{commit}/{file_path:path}")
async def file_head_proxy(org: str, repo: str, commit: str, file_path: str, request: Request, repo_type: str = "model"):
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return Response(content="This repository is forbidden by the mirror. ", status_code=403)
    if not await check_commit_hf(app, repo_type, org, repo, commit):
        return Response(content="This repository is not accessible. ", status_code=404)
    generator = file_head_generator(app, repo_type, org, repo, commit, file_path, request)
    headers = await generator.__anext__()
    return StreamingResponse(generator, headers=headers)

@app.get("/{repo_type}s/{org}/{repo}/resolve/{commit}/{file_path:path}")
@app.get("/{org}/{repo}/resolve/{commit}/{file_path:path}")
async def file_proxy(org: str, repo: str, commit: str, file_path: str, request: Request, repo_type: str = "model"):
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return Response(content="This repository is forbidden by the mirror. ", status_code=403)
    if not await check_commit_hf(app, repo_type, org, repo, commit):
        return Response(content="This repository is not accessible. ", status_code=404)
    generator = file_get_generator(app, repo_type, org, repo, commit, file_path, request)
    headers = await generator.__anext__()
    return StreamingResponse(generator, headers=headers)

@app.get("/repos/{dir1}/{dir2}/{hash_repo}/{hash_file}")
async def lfs_proxy(dir1: str, dir2: str, hash_repo: str, hash_file: str, request: Request):
    repo_type = "model"
    lfs_url = f"{app.app_settings.hf_lfs_url}/repos/{dir1}/{dir2}/{hash_repo}/{hash_file}"
    save_path = f"{dir1}/{dir2}/{hash_repo}/{hash_file}"
    generator = lfs_get_generator(app, repo_type, lfs_url, save_path, request)
    headers = await generator.__anext__()
    return StreamingResponse(generator, headers=headers)

@app.get("/datasets/hendrycks_test/{hash_file}")
async def lfs_proxy(hash_file: str, request: Request):
    repo_type = "dataset"
    lfs_url = f"{app.app_settings.hf_lfs_url}/datasets/hendrycks_test/{hash_file}"
    save_path = f"hendrycks_test/{hash_file}"
    generator = lfs_get_generator(app, repo_type, lfs_url, save_path, request)
    headers = await generator.__anext__()
    return StreamingResponse(generator, headers=headers)

if __name__ in ["__main__", "olah.server"]:
    parser = argparse.ArgumentParser(
        description="Olah Huggingface Mirror Server."
    )
    parser.add_argument("--config", "-c", type=str, default="")
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=8090)
    parser.add_argument("--ssl-key", type=str, default=None)
    parser.add_argument("--ssl-cert", type=str, default=None)
    parser.add_argument("--repos-path", type=str, default="./repos")
    parser.add_argument("--hf-url", type=str, default="https://huggingface.co")
    parser.add_argument("--hf-lfs-url", type=str, default="https://cdn-lfs.huggingface.co")
    parser.add_argument("--mirror-url", type=str, default="http://localhost:8090")
    parser.add_argument("--mirror-lfs-url", type=str, default="http://localhost:8090")
    args = parser.parse_args()
    
    def is_default_value(args, arg_name):
        if hasattr(args, arg_name):
            arg_value = getattr(args, arg_name)
            arg_default = parser.get_default(arg_name)
            return arg_value == arg_default
        return False

    if args.config != "":
        config = OlahConfig(args.config)
    else:
        config = OlahConfig()
    
    if is_default_value(args, "host"):
        args.host = config.host
    if is_default_value(args, "port"):
        args.port = config.port
    if is_default_value(args, "ssl_key"):
        args.ssl_key = config.ssl_key
    if is_default_value(args, "ssl_cert"):
        args.ssl_cert = config.ssl_cert
    if is_default_value(args, "repos_path"):
        args.repos_path = config.repos_path
    if is_default_value(args, "hf_url"):
        args.hf_url = config.hf_url
    if is_default_value(args, "hf_lfs_url"):
        args.hf_lfs_url = config.hf_lfs_url
    if is_default_value(args, "mirror_url"):
        args.mirror_url = config.mirror_url
    if is_default_value(args, "mirror_lfs_url"):
        args.mirror_lfs_url = config.mirror_lfs_url

    app.app_settings = AppSettings(
        config=config,
        repos_path=args.repos_path,
        hf_url=args.hf_url,
        hf_lfs_url=args.hf_lfs_url,
        mirror_url=args.mirror_url,
        mirror_lfs_url=args.mirror_lfs_url,
    )

    import uvicorn
    if __name__ == "__main__":
        uvicorn.run(
            "olah.server:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=False,
            ssl_keyfile=args.ssl_key,
            ssl_certfile=args.ssl_cert
        )



import datetime
import json
import os
import tempfile
import shutil
from fastapi import FastAPI, Header, Request
import httpx
import pytz

from olah.constants import CHUNK_SIZE, LFS_FILE_BLOCK, WORKER_API_TIMEOUT


async def lfs_get_generator(app, repo_type: str, lfs_url: str, save_path: str, request: Request):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    # save
    repos_path = app.app_settings.repos_path
    save_dir = os.path.join(repos_path, f"lfs/{repo_type}s/{save_path}")
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)
    
    # lfs meta
    lfs_meta_path = os.path.join(save_dir, "meta.json")
    if os.path.exists(lfs_meta_path):
        with open(lfs_meta_path, "r", encoding="utf-8") as f:
            lfs_meta = json.loads(f.read())
    else:
        async with httpx.AsyncClient() as client:
            async with client.stream(
                method="GET", url=lfs_url,
                headers={"range": "-"},
                params=request.query_params,
                timeout=WORKER_API_TIMEOUT,
            ) as response:
                file_size = response.headers["content-length"]
                req_headers = {k: v for k, v in response.headers.items()}
        lfs_meta = {
            "lfs_file_block": LFS_FILE_BLOCK,
            "file_size": int(file_size),
            "req_headers": req_headers,
        }
        with open(lfs_meta_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(lfs_meta))
    # range
    file_size = lfs_meta["file_size"]
    if "range" in headers:
        file_range = headers['range'] # 'bytes=1887436800-'
        if file_range.startswith("bytes="):
            file_range = file_range[6:]
        start_pos, end_pos = file_range.split("-")
        if len(start_pos) != 0:
            start_pos = int(start_pos)
        else:
            start_pos = 0
        if len(end_pos) != 0:
            end_pos = int(end_pos)
        else:
            end_pos = file_size
    else:
        start_pos = 0
        end_pos = file_size

    # block
    lfs_file_block = lfs_meta["lfs_file_block"]
    start_block = start_pos // lfs_file_block
    end_block = end_pos // lfs_file_block

    new_headers = lfs_meta["req_headers"]
    new_headers["date"] = datetime.datetime.now(pytz.timezone('GMT')).strftime('%a, %d %b %Y %H:%M:%S %Z')
    new_headers["content-length"] = str(end_pos - start_pos)

    yield new_headers
    cur_pos = start_pos
    cur_block = start_block

    while cur_block <= end_block:
        save_path = os.path.join(save_dir, f"block-{cur_block}.bin")
        use_cache = os.path.exists(save_path)
        block_start_pos = cur_block * lfs_file_block
        block_end_pos = min((cur_block + 1) * lfs_file_block, file_size)

        # proxy
        if use_cache:
            with open(save_path, "rb") as f:
                sub_chunk_start_pos = block_start_pos
                while True:
                    raw_chunk = f.read(CHUNK_SIZE)
                    if not raw_chunk:
                        break

                    chunk = raw_chunk
                    if cur_pos >= sub_chunk_start_pos and cur_pos < sub_chunk_start_pos + len(raw_chunk):
                        chunk = chunk[cur_pos - sub_chunk_start_pos:]
                    elif cur_pos >= sub_chunk_start_pos + len(raw_chunk):
                        chunk = bytes([])
                    elif cur_pos < sub_chunk_start_pos:
                        pass

                    if cur_pos + len(chunk) > block_end_pos:
                        chunk = chunk[:-(cur_pos + len(chunk) - block_end_pos)]
                        print("Warning: This maybe a bug, sending chunk is larger than content length.")
                    
                    if len(chunk) != 0:
                        yield chunk
                    cur_pos += len(chunk)
                    sub_chunk_start_pos += len(raw_chunk)
        else:
            try:
                temp_file_path = None
                async with httpx.AsyncClient() as client:
                    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as temp_file:
                        headers["range"] = f"bytes={block_start_pos}-{block_end_pos - 1}"
                        async with client.stream(
                            method="GET", url=lfs_url,
                            headers=headers,
                            params=request.query_params,
                            timeout=WORKER_API_TIMEOUT,
                        ) as response:
                            raw_bytes = 0
                            sub_chunk_start_pos = block_start_pos
                            async for raw_chunk in response.aiter_raw():
                                if not raw_chunk:
                                    continue
                                temp_file.write(raw_chunk)

                                stream_chunk = raw_chunk
                                
                                if cur_pos > sub_chunk_start_pos and cur_pos < sub_chunk_start_pos + len(raw_chunk):
                                    stream_chunk = stream_chunk[cur_pos - sub_chunk_start_pos:]
                                elif cur_pos >= sub_chunk_start_pos + len(raw_chunk):
                                    stream_chunk = bytes([])
                                elif cur_pos < sub_chunk_start_pos:
                                    pass

                                if cur_pos + len(stream_chunk) > block_end_pos:
                                    stream_chunk = stream_chunk[:-(cur_pos + len(stream_chunk) - block_end_pos)]
                                    print("Warning: This maybe a bug, sending chunk is larger than content length.")

                                if len(stream_chunk) != 0:
                                    yield stream_chunk
                                cur_pos += len(stream_chunk)
                                raw_bytes += len(raw_chunk)
                                sub_chunk_start_pos += len(raw_chunk)
                                if raw_bytes >= block_end_pos - block_start_pos:
                                    break
                        temp_file_path = temp_file.name
                    shutil.copyfile(temp_file_path, save_path)
            finally:
                if temp_file_path is not None:
                    os.remove(temp_file_path)
        cur_block += 1




from typing import Literal

import httpx
from olah.configs import OlahConfig
from olah.constants import WORKER_API_TIMEOUT

async def check_commit_hf(app, repo_type: Literal["model", "dataset"], org: str, repo: str, commit: str) -> bool:
    async with httpx.AsyncClient() as client:
        response = await client.get(f"{app.app_settings.hf_url}/api/{repo_type}s/{org}/{repo}/revision/{commit}",
                   timeout=WORKER_API_TIMEOUT)
    return response.status_code == 200

async def check_proxy_rules_hf(app, repo_type: Literal["model", "dataset"], org: str, repo: str) -> bool:
    config: OlahConfig = app.app_settings.config
    return config.proxy.allow(f"{org}/{repo}")

async def check_cache_rules_hf(app, repo_type: Literal["model", "dataset"], org: str, repo: str) -> bool:
    config: OlahConfig = app.app_settings.config
    return config.cache.allow(f"{org}/{repo}")
